<!doctype html>

<html>

<head>
  <title>{{title}}</title>
  <link rel='stylesheet' href='/css/style.css'>
</head>

<body>
  body: {{{body}}}

  <hr/>
  post body
  <hr/>

  {{> partial5 }}
</body>

</html>
